# zyx ===============================================

from ._loader import UtilLazyLoader


class BaseModel(UtilLazyLoader):
    pass


BaseModel.init("pydantic.main", "BaseModel")


class Field(UtilLazyLoader):
    pass


Field.init("pydantic.fields", "Field")


class logger(UtilLazyLoader):
    pass


logger.init("loguru", "logger")


class console(UtilLazyLoader):
    pass


console.init("rich.console", "Console")


class tqdm(UtilLazyLoader):
    pass


tqdm.init("tqdm", "tqdm")


class tqdm_notebook(UtilLazyLoader):
    pass


tqdm_notebook.init("tqdm.notebook", "tqdm_notebook")


class lightning(UtilLazyLoader):
    pass


lightning.init("zyx.functions.lightning", "lightning")

# ====================================================



# zyx ==============================================================================

from .ext._loader import UtilLazyLoader


class tool(UtilLazyLoader):
    pass


tool.init("langchain_core.tools", "tool")


class ChatPromptTemplate(UtilLazyLoader):
    pass


ChatPromptTemplate.init("langchain_core.prompts", "ChatPromptTemplate")


class MessagesPlaceholder(UtilLazyLoader):
    pass


MessagesPlaceholder.init("langchain_core.prompts", "MessagesPlaceholder")


class BaseMessage(UtilLazyLoader):
    pass


BaseMessage.init("langchain_core.messages", "BaseMessage")


class AIMessage(UtilLazyLoader):
    pass


AIMessage.init("langchain_core.messages", "AIMessage")


class HumanMessage(UtilLazyLoader):
    pass


HumanMessage.init("langchain_core.messages", "HumanMessage")


class SystemMessage(UtilLazyLoader):
    pass


SystemMessage.init("langchain_core.messages", "SystemMessage")


class ToolMessage(UtilLazyLoader):
    pass


ToolMessage.init("langchain_core.messages", "ToolMessage")


class BaseModel(UtilLazyLoader):
    pass


BaseModel.init("langchain_core.pydantic_v1", "BaseModel")


class Field(UtilLazyLoader):
    pass


Field.init("langchain_core.pydantic_v1", "Field")


class ensure_config(UtilLazyLoader):
    pass


ensure_config.init("langchain_core.runnables.config", "ensure_config")


class ChatOpenAI(UtilLazyLoader):
    pass


ChatOpenAI.init("langchain_openai.chat_models.base", "ChatOpenAI")


class OpenAIEmbeddings(UtilLazyLoader):
    pass


OpenAIEmbeddings.init("langchain_openai.embeddings.base", "OpenAIEmbeddings")


class ChatAnthropic(UtilLazyLoader):
    pass


ChatAnthropic.init("langchain_anthropic.chat_models", "ChatAnthropic")



# zyx ==============================================================================

from langchain_core.tools import tool as tool_type
from langchain_core.prompts import ChatPromptTemplate as ChatPromptTemplate_type
from langchain_core.prompts import MessagesPlaceholder as MessagesPlaceholder_type
from langchain_core.messages import BaseMessage as BaseMessage_type
from langchain_core.messages import AIMessage as AIMessage_type
from langchain_core.messages import HumanMessage as HumanMessage_type
from langchain_core.messages import SystemMessage as SystemMessage_type
from langchain_core.messages import ToolMessage as ToolMessage_type
from langchain_core.pydantic_v1 import BaseModel as BaseModel_type
from langchain_core.pydantic_v1 import Field as Field_type
from langchain_core.runnables.config import ensure_config as ensure_config_type
from langchain_openai.chat_models.base import ChatOpenAI as ChatOpenAI_type
from langchain_openai.embeddings.base import OpenAIEmbeddings as OpenAIEmbeddings_type
from langchain_anthropic.chat_models import ChatAnthropic as ChatAnthropic_type

tool = tool_type
ChatPromptTemplate = ChatPromptTemplate_type
MessagesPlaceholder = MessagesPlaceholder_type
BaseMessage = BaseMessage_type
AIMessage = AIMessage_type
HumanMessage = HumanMessage_type
SystemMessage = SystemMessage_type
ToolMessage = ToolMessage_type
BaseModel = BaseModel_type
Field = Field_type
ensure_config = ensure_config_type
ChatOpenAI = ChatOpenAI_type
OpenAIEmbeddings = OpenAIEmbeddings_type
ChatAnthropic = ChatAnthropic_type



# zyx ==============================================================================

from ._loader import UtilLazyLoader


class db(UtilLazyLoader):
    pass


db.init("sqlite3", "sqlite3")


class memory(UtilLazyLoader):
    pass


memory.init("mem0.memory.main", "Memory")


class qdrant(UtilLazyLoader):
    pass


qdrant.init("qdrant_client.qdrant_client", "QdrantClient")



# zyx ==============================================================================

from ._loader import UtilLazyLoader


class batch_completion(UtilLazyLoader):
    pass


batch_completion.init("litellm.main", "batch_completion_models_all_responses")


class completion(UtilLazyLoader):
    pass


completion.init("zyx.functions.completion", "completion")


class embedding(UtilLazyLoader):
    pass


embedding.init("litellm.main", "embedding")


class Inference(UtilLazyLoader):
    pass


Inference.init("huggingface_hub.inference._client", "InferenceClient")


class ollama_embedding(UtilLazyLoader):
    pass


ollama_embedding.init("ollama", "embeddings")



# zyx ==============================================================================

from langgraph.prebuilt.chat_agent_executor import (
    create_react_agent as create_react_agent_type,
)
from langgraph.checkpoint.memory import MemorySaver as MemorySaver_type
from langgraph.checkpoint.sqlite import SqliteSaver as SqliteSaver_type
from langgraph.constants import Send as Send_type
from langgraph.graph import START as START_type
from langgraph.graph import END as END_type
from langgraph.graph.message import MessageGraph as MessageGraph_type
from langgraph.graph.state import StateGraph as StateGraph_type

react_agent = create_react_agent_type
MemorySaver = MemorySaver_type
SqliteSaver = SqliteSaver_type
Send = Send_type
START = START_type
END = END_type
MessageGraph = MessageGraph_type
StateGraph = StateGraph_type



# zyx ==============================================================================

__all__ = [
    "batch_completion",
    "completion",
    "embedding",
    "Inference",
    "ollama_embedding",
    "db",
    "Memory",
    "Qdrant",
    "logger",
    "console",
    "BaseModel",
    "Field",
    "tqdm",
    "tqdm_notebook",
]

from .ext.ai import batch_completion as batch_completion
from .ext.ai import completion as completion
from .ext.ai import embedding as embedding
from .ext.ai import Inference as Inference
from .ext.ai import ollama_embedding as ollama_embedding
from .ext.data import db as db
from .ext.data import memory as Memory
from .ext.data import qdrant as Qdrant
from .ext.util import logger as logger
from .ext.util import console as console
from .ext.util import BaseModel as BaseModel
from .ext.util import Field as Field
from .ext.util import tqdm as tqdm
from .ext.util import tqdm_notebook as tqdm_notebook

from rich import print
import builtins as __builtins__

__builtins__.print = print



# zyx ==============================================================================

from litellm.main import (
    batch_completion_models_all_responses as batch_completion_models_all_responses_type,
)
from .functions.completion import completion as completion_type
from .functions.lightning import lightning as lightning_type
from litellm.main import embedding as embedding_type
from huggingface_hub.inference._client import InferenceClient as InferenceClient_type
from ollama import embeddings as ollama_embedding_type
import sqlite3 as sqlite3_type
from mem0.memory.main import Memory as Memory_type
from qdrant_client.qdrant_client import QdrantClient as QdrantClient_type
from loguru import logger as logger_type
from pydantic.main import BaseModel as BaseModel_type
from pydantic.fields import Field as Field_type
from rich.console import Console as console_type
from tqdm import tqdm as tqdm_type
from tqdm.notebook import tqdm_notebook as tqdm_notebook_type

# ---------------------------------------------------

batch_completion = batch_completion_models_all_responses_type
completion = completion_type
lightning = lightning_type

# ---------------------------------------------------

embedding = embedding_type
Inference = InferenceClient_type
ollama_embedding = ollama_embedding_type

# ---------------------------------------------------

db = sqlite3_type
memory = Memory_type
qdrant = QdrantClient_type

# ---------------------------------------------------

logger = logger_type
console = console_type
BaseModel = BaseModel_type
Field = Field_type
tqdm = tqdm_type
tqdm_notebook = tqdm_notebook_type

# ---------------------------------------------------



# zyx ==============================================================================

import concurrent.futures
import functools
import psutil
import time
from typing import Callable, Iterable, Optional, Any, TypeVar, List, Generator

T = TypeVar("T")
R = TypeVar("R")


def lightning(
    batch_size: Optional[int] = None,
    verbose: bool = False,
    timeout: Optional[float] = None,
) -> Callable[[Callable[[T], R]], Callable[[Iterable[T]], List[R]]]:
    """
    A decorator for high-performance parallel processing of CPU-bound tasks.

    This decorator can be used to automatically parallelize the execution of a function
    over an iterable input (including generators), with optional batching for improved performance.

    Args:
        batch_size (Optional[int]): The size of each batch. If None, no batching is performed.
        verbose (bool): Whether to print execution time and memory usage information.
        timeout (Optional[float]): The maximum number of seconds to wait for all tasks to complete.

    Returns:
        Callable: A decorator function that wraps the original function with parallel processing capabilities.

    Example:
        @zyx.lightning(batch_size=10, verbose=True)
        def square(x: int) -> int:
            return x * x

        results = square(range(100))
    """

    def decorator(func: Callable[[T], R]) -> Callable[[Iterable[T]], List[R]]:
        @functools.wraps(func)
        def wrapper(iterable: Iterable[T], *args: Any, **kwargs: Any) -> List[R]:
            num_workers = psutil.cpu_count(logical=False) or 4

            def process_batch(batch: List[T]) -> List[R]:
                return [func(item, *args, **kwargs) for item in batch]

            results: List[R] = []
            start_time = time.time() if verbose else None
            start_memory = psutil.virtual_memory().used if verbose else None

            # Convert generator to list if batching is required
            if batch_size and isinstance(iterable, Generator):
                iterable = list(iterable)

            with concurrent.futures.ThreadPoolExecutor(
                max_workers=num_workers
            ) as executor:
                if batch_size:
                    batches = [
                        list(iterable)[i : i + batch_size]
                        for i in range(0, len(list(iterable)), batch_size)
                    ]
                    future_to_batch = {
                        executor.submit(process_batch, batch): batch
                        for batch in batches
                    }
                else:
                    future_to_item = {
                        executor.submit(func, item, *args, **kwargs): item
                        for item in iterable
                    }

                for future in concurrent.futures.as_completed(
                    future_to_batch if batch_size else future_to_item, timeout=timeout
                ):
                    try:
                        batch_results = future.result()
                        results.extend(batch_results if batch_size else [batch_results])
                    except Exception as exc:
                        item = (
                            future_to_batch[future]
                            if batch_size
                            else future_to_item[future]
                        )
                        print(f"Generated an exception: {exc} with item/batch: {item}")

            if verbose:
                end_time = time.time()
                end_memory = psutil.virtual_memory().used
                execution_time = end_time - start_time if start_time else 0
                memory_used = end_memory - start_memory if start_memory else 0
                print(f"Execution Time: {execution_time:.2f} seconds")
                print(f"Memory Used: {memory_used} bytes")
                print(f"Worker Threads Used: {num_workers}")
                print(f"Available Memory: {psutil.virtual_memory().available} bytes")
                print(f"Total Memory: {psutil.virtual_memory().total} bytes")

            return results

        return wrapper

    return decorator


# ==============================================================================



# zyx ==============================================================================

__all__ = [
    "react_agent",
    "MemorySaver",
    "SqliteSaver",
    "Send",
    "START",
    "END",
    "MessageGraph",
    "StateGraph",
]


from .ext._loader import UtilLazyLoader


class react_agent(UtilLazyLoader):
    pass


react_agent.init("langgraph.prebuilt.chat_agent_executor", "create_react_agent")


class MemorySaver(UtilLazyLoader):
    pass


MemorySaver.init("langgraph.checkpoint.memory", "MemorySaver")


class SqliteSaver(UtilLazyLoader):
    pass


SqliteSaver.init("langgraph.checkpoint.sqlite", "SqliteSaver")


class Send(UtilLazyLoader):
    pass


Send.init("langgraph.constants", "Send")


class START(UtilLazyLoader):
    pass


START.init("langgraph.graph", "START")


class END(UtilLazyLoader):
    pass


END.init("langgraph.graph", "END")


class MessageGraph(UtilLazyLoader):
    pass


MessageGraph.init("langgraph.graph.message", "MessageGraph")


class StateGraph(UtilLazyLoader):
    pass


StateGraph.init("langgraph.graph.state", "StateGraph")



# zyx ==============================================================================

from typing import Type, Optional, Union

BaseModel = Type["BaseModel"]
ModelResponse = Type["ModelResponse"]


# ==============================================================================
def completion(
    messages: Union[str, list[str]],
    model: Optional[str] = "openai/gpt-3.5-turbo",
    response_model: Type["BaseModel"] = None,  # Forward reference for BaseModel
    base_url: Optional[str] = None,
    temperature: Optional[float] = 0.5,
    max_tokens: Optional[int] = None,
    max_retries: Optional[int] = 3,
    strict: Optional[bool] = True,
    debug: Optional[bool] = False,
    **kwargs,
) -> Union["ModelResponse", "BaseModel"]:
    """
    Runs a litellm completion, tied in with the Instructor framework.

    Parameters:
        messages (Union[str, list[str]]) : The message(s) to send to the model
        model (str) : The model to use for completion
        response_model (Type['BaseModel']) : The response model to use
        base_url (Optional[str]) : The base url for the instructor
        temperature (Optional[float]) : The temperature for the completion
        max_tokens (Optional[int]) : The maximum tokens to use for the completion
        max_retries (Optional[int]) : The maximum retries to use for the completion
        strict (bool) : Whether to use strict mode for the completion

    Returns:
        Union['ModelResponse', 'BaseModel'] : The response from the completion
    """

    if debug:
        from ..ext.util import logger

        logger = logger

    if not messages:
        if debug:
            logger.error("No messages provided")
            return None
        else:
            raise ValueError("No messages provided")

    if isinstance(messages, str):
        messages = [{"role": "user", "content": messages}]

    if not response_model:
        from litellm.main import completion as litellm_completion

        response = litellm_completion(
            model=model,
            messages=messages,
            base_url=base_url,
            temperature=temperature,
            max_tokens=max_tokens,
            max_retries=max_retries,
            **kwargs,
        )
        return _return_response(response)

    if model.startswith("ollama/"):
        return _ollama_instruct(
            model,
            messages,
            response_model,
            temperature=temperature,
            base_url=base_url,
            max_tokens=max_tokens,
            max_retries=max_retries,
            strict=strict,
            **kwargs,
        )
    else:
        return _litellm_instruct(
            model,
            messages,
            response_model,
            temperature=temperature,
            base_url=base_url,
            max_tokens=max_tokens,
            max_retries=max_retries,
            strict=strict,
            **kwargs,
        )


# ==============================================================================
def _return_response(response: "ModelResponse" = None) -> "ModelResponse":
    """
    Returns the response from the completion function

    Parameters:
        response : The response from the completion function
    """
    return response


# ==============================================================================
def _ollama_instruct(
    model: str,
    messages: list[dict],
    response_model: Type["BaseModel"],
    base_url: Optional[str] = None,
    temperature: Optional[float] = 0.5,
    max_tokens: Optional[int] = None,
    max_retries: Optional[int] = 3,
    strict: bool = True,
    **kwargs,
):
    """
    Runs the ollama instructor function

    Parameters:
        model (str) : The model to use for completion
        messages (list[dict]) : A list of messages to send to the model
        response_model (Type['BaseModel']) : The response model to use
        temperature (float) : The temperature to use for completion
        max_tokens (int) : The maximum number of tokens to generate
        max_retries (int) : The maximum number of retries to attempt
        strict (bool) : Whether to raise an exception on failure
        **kwargs : Additional arguments to pass to the completion function
    """
    if not base_url:
        base_url = "http://localhost:11434/v1"

    from instructor.client import from_openai as FromOpenAI
    from instructor.mode import Mode
    from openai import OpenAI

    client = FromOpenAI.from_openai(
        OpenAI(
            base_url=base_url,
            api_key="ollama",
        ),
        mode=Mode.JSON,
    )
    resp = client.chat.completions.create(
        model=model.split("/")[1],  # Remove the Ollama prefix
        messages=messages,
        response_model=response_model,
        temperature=temperature,
        max_tokens=max_tokens,
        max_retries=max_retries,
        strict=strict,
        **kwargs,
    )
    return resp


# -----------------------------------------------------------------------------
def _litellm_instruct(
    model: str,
    messages: list[dict],
    response_model: Type["BaseModel"],
    base_url: Optional[str] = None,
    temperature: Optional[float] = 0.5,
    max_tokens: Optional[int] = None,
    max_retries: Optional[int] = 3,
    strict: bool = True,
    **kwargs,
):
    """
    Runs the litellm instructor function

    Parameters:
        model (str) : The model to use for completion
        messages (list[dict]) : A list of messages to send to the model
        response_model (Type['BaseModel']) : The response model to use
        temperature (float) : The temperature to use for completion
        max_tokens (int) : The maximum number of tokens to generate
        max_retries (int) : The maximum number of retries to attempt
        strict (bool) : Whether to raise an exception on failure
        **kwargs : Additional arguments to pass to the completion function
    """
    from litellm.main import completion as litellm_completion
    from instructor.client import from_litellm

    client = from_litellm(litellm_completion)
    response = client.chat.completions.create(
        model=model,
        messages=messages,
        response_model=response_model,
        temperature=temperature,
        base_url=base_url,
        max_tokens=max_tokens,
        max_retries=max_retries,
        strict=strict,
        **kwargs,
    )
    return response


# ==============================================================================

# ==============================================================================
if __name__ == "__main__":
    print(completion("How are you?"))



# zyx ===============================================

from importlib import import_module
from types import ModuleType
from typing import Any, Optional
import threading


class LazyLoaderMeta(type):
    """
    Meta class for lazy loading attributes from a module.
    """

    def __getattr__(cls, name: str) -> Any:
        module = cls._load()
        try:
            return getattr(module, name)
        except AttributeError as e:
            raise e

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        module = cls._load()
        return module(*args, **kwargs)


class UtilLazyLoader(metaclass=LazyLoaderMeta):
    """
    Lazy loader for a module.
    """

    _module_cache: Optional[ModuleType] = None
    _lock = threading.Lock()

    @classmethod
    def _load(cls) -> ModuleType:
        if cls._module_cache is None:
            with cls._lock:
                if cls._module_cache is None:
                    try:
                        module = import_module(cls._module_name)
                        cls._module_cache = getattr(module, cls._attr_name)
                    except (ImportError, AttributeError) as e:
                        raise e
        return cls._module_cache

    @classmethod
    def init(cls, module_name: str, attr_name: str) -> None:
        cls._module_name = module_name
        cls._attr_name = attr_name


# ==============================================================================
